#ifndef _BASE_H_
#define _BASE_H_

#include <sys/types.h>
#include <sys/un.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "buffer.h"
#include "array.h"
#include "chunk.h"
#include "keyvalue.h"
#include "settings.h"

#include "config.h"

/* select event-system */

#if defined(HAVE_EPOLL_CTL) && defined(HAVE_SYS_EPOLL_H)
# define USE_LINUX_EPOLL
# include <sys/epoll.h>
#endif

#if defined HAVE_POLL && defined(HAVE_SYS_POLL_H)
# define USE_POLL
# include <sys/poll.h>
# if defined HAVE_SIGTIMEDWAIT && defined(__linux__)
#  define USE_LINUX_SIGIO
#  include <signal.h>
# endif
#endif

#if defined HAVE_SELECT
# define USE_SELECT
# include <sys/select.h>
#endif

#if defined HAVE_SYS_DEVPOLL_H && defined(__sun)
# define USE_SOLARIS_DEVPOLL
# include <sys/devpoll.h>
#endif

#if defined HAVE_SYS_EVENT_H && defined HAVE_KQUEUE
# define USE_FREEBSD_KQUEUE
# include <sys/event.h>
#endif

#if defined HAVE_LIBSSL && defined HAVE_OPENSSL_SSL_H
# define USE_OPENSSL
# include <openssl/ssl.h> 
#endif

#ifndef O_LARGEFILE
# define O_LARGEFILE 0
#endif

#include "settings.h"

typedef enum { T_CONFIG_UNSET, 
		T_CONFIG_STRING, 
		T_CONFIG_SHORT, 
		T_CONFIG_BOOLEAN, 
		T_CONFIG_ARRAY, 
		T_CONFIG_LOCAL, 
		T_CONFIG_DEPRECATED
} config_values_type_t;

typedef struct {
	const char *key;
	void *destination;
	
	config_values_type_t type;
} config_values_t;

#define FDEVENT_IN     BV(0)
#define FDEVENT_PRI    BV(1)
#define FDEVENT_OUT    BV(2)
#define FDEVENT_ERR    BV(3)
#define FDEVENT_HUP    BV(4)
#define FDEVENT_NVAL   BV(5)

typedef enum { FD_EVENT_TYPE_UNSET = -1, 
		FD_EVENT_TYPE_CONNECTION, 
		FD_EVENT_TYPE_FCGI_CONNECTION, 
		FD_EVENT_TYPE_DIRWATCH, 
		FD_EVENT_TYPE_CGI_CONNECTION 
} fd_event_t;

typedef enum { FDEVENT_HANDLER_UNSET, 
		FDEVENT_HANDLER_SELECT,
		FDEVENT_HANDLER_POLL,
		FDEVENT_HANDLER_LINUX_RTSIG,
		FDEVENT_HANDLER_LINUX_SYSEPOLL,
		FDEVENT_HANDLER_SOLARIS_DEVPOLL,
		FDEVENT_HANDLER_FREEBSD_KQUEUE
} fdevent_handler_t;

typedef enum { HANDLER_UNSET, 
		HANDLER_GO_ON, 
		HANDLER_FINISHED,
		HANDLER_COMEBACK, 
		HANDLER_WAIT_FOR_EVENT, 
		HANDLER_ERROR 
} handler_t;

/**
 * a mapping from fd to connection structure
 * 
 */
typedef struct {
	int fd;                  /**< the fd */
	void *conn;              /**< a reference the corresponding data-structure */
	fd_event_t fd_type;      /**< type of the fd */
	int events;              /**< registered events */
	int revents;
} fd_conn;

typedef struct {
	fd_conn *ptr;
	
	size_t size;
	size_t used;
} fd_conn_buffer;

/**
 * array of unused fd's
 * 
 */

typedef struct {
	size_t ndx;
} fd_unused;

typedef struct {
	fd_unused *ptr;
	
	size_t size;
	size_t used;
} fd_unused_array;


/**
 * fd-event handler for select(), poll() and rt-signals on Linux 2.4
 * 
 */
typedef struct fdevents {
	fdevent_handler_t type;
	
#ifdef USE_LINUX_SIGIO
	int in_sigio;
	int signum;
	sigset_t sigset;
	siginfo_t siginfo;
#endif
#ifdef USE_LINUX_EPOLL
	int epoll_fd;
	struct epoll_event epoll_events[MAX_EPOLL_EVENTS];
#endif
#ifdef USE_POLL
	struct pollfd *pollfds;
	
	size_t size;
	size_t used;
	
	fd_unused_array unused;
#endif
#ifdef USE_SELECT
	fd_set select_read;
	fd_set select_write;
	fd_set select_error;
	
	fd_set select_set_read;
	fd_set select_set_write;
	fd_set select_set_error;
	
	int max_fd;
#endif
#ifdef USE_SOLARIS_DEVPOLL
	int devpoll_fd;
	struct pollfd *devpollfds;
#endif
#ifdef USE_FREEBSD_KQUEUE
	int kqueue_fd;
#endif
	fd_conn_buffer *fdc;
	int fdc_changed;
	
	int (*reset)(struct fdevents *ev);
	void (*free)(struct fdevents *ev);
	
	int (*event_add)(struct fdevents *ev, int fde_ndx, int fd, int events);
	int (*event_del)(struct fdevents *ev, int fde_ndx, int fd);
	int (*event_get_revent)(struct fdevents *ev, size_t ndx);
	int (*event_get_fd)(struct fdevents *ev, size_t ndx);
	int (*event_next_fdndx)(struct fdevents *ev, int ndx);
	
	int (*poll)(struct fdevents *ev, int timeout_ms);
	
	int (*fcntl_set)(struct fdevents *ev, int fd);
	
	
} fdevents;

typedef enum { DIRECT, CGI, FASTCGI, MODULE } connection_type;

typedef struct {
	char *key;
	connection_type type;
	char *value;
} request_handler;

typedef struct {
	char *key;
	char *host;
	unsigned short port;
	int used;
	short factor;
} fcgi_connections;


typedef union {
#ifdef HAVE_IPV6
	struct sockaddr_in6 ipv6;
#endif
	struct sockaddr_in ipv4;
	struct sockaddr plain;
} sock_addr;

/* fcgi_response_header contains ... */
#define HTTP_STATUS         BV(0)
#define HTTP_CONNECTION     BV(1)
#define HTTP_CONTENT_LENGTH BV(2)
#define HTTP_DATE           BV(3)
#define HTTP_LOCATION       BV(4)

typedef struct {
	/** HEADER */
	/* the request-line */
	buffer *request;
	buffer *uri;
	
	http_method_t  http_method;
	http_version_t http_version;
	
	buffer *request_line;
	
	/* strings to the header */
	buffer *http_host; /* not alloced */
	char   *http_range;
	char   *http_content_type;
	char   *http_if_modified_since;
	char   *http_if_none_match;
	
	array  *headers;
	
	/* CONTENT */
	buffer *content;
	size_t content_length; /* returned by strtoul() */
	
	/* internal representation */
	int     accept_encoding;
	
	/* internal */
	buffer *pathinfo;
} request;

typedef struct {
	off_t  content_length;
	int     keep_alive;
	
	array  *headers;
	
	enum { 
		HTTP_TRANSFER_ENCODING_IDENTITY, HTTP_TRANSFER_ENCODING_CHUNKED
	} transfer_encoding;
} response;

typedef enum { FCGI_STATE_INIT, FCGI_STATE_CONNECT, FCGI_STATE_PREPARE_WRITE, FCGI_STATE_WRITE, FCGI_STATE_READ, FCGI_STATE_ERROR } fcgi_connection_state_t;
typedef struct {
	buffer *response;
	size_t response_len;
	int response_type;
	int response_padding;
	size_t response_request_id;
	
	data_fastcgi *host;
	
	fcgi_connection_state_t state;
	time_t state_timestamp;
	
	buffer *write_buffer;
	size_t  write_offset;
	
	read_buffer *rb;
	
	buffer *response_header;

	size_t request_id;
	int fd; /* fd to the fastcgi process */
	int fde_ndx; /* index into the fd-event buffer */

	size_t path_info_offset; /* start of path_info in uri.path */
} fcgi;

typedef struct {
	pid_t pid;
	int fd;
	int fde_ndx; /* index into the fd-event buffer */
	
	buffer *response;
	buffer *response_header;
} cgi;

typedef struct {
	buffer *basedir;
	
	buffer *trigger_handler;
	
	buffer *session_id;
	
	buffer_array *eval;
	buffer_array *trigger_if;
	buffer_array *output_include;
} cache;

typedef struct {
	buffer *name;
	buffer *etag;
	
	struct stat st;
	
	int    fd;
	int    fde_ndx;
	char   *mmap_p;
	
	size_t in_use;
	size_t is_dirty;
	
	time_t stat_ts;
} file_cache_entry;

typedef struct {
	buffer *scheme;
	buffer *authority;
	buffer *path;
	buffer *path_raw;
	buffer *query;
} request_uri;

typedef struct {
	buffer *path;
	buffer *doc_root;
	buffer *etag;
} physical;

typedef struct {
	file_cache_entry **ptr;
	
	size_t size;
	size_t used;
} file_cache;

typedef enum { CONNECT, REQUEST_START, READ, REQUEST_END, READ_POST, HANDLE_REQUEST, RESPONSE_START, WRITE, RESPONSE_END, ERROR, CLOSE } connection_state_t;

typedef struct {
	connection_state_t state;
	
	time_t read_idle_ts;
	time_t close_timeout_ts;
	time_t write_request_ts;
	
	time_t connection_start;
	time_t request_start;
	
	struct timeval start_tv;
	
	int fd;
	int fde_ndx;
	int ndx;
	
	/* fd states */
	int is_readable;
	int is_writable;
	
	enum { FALSE, TRUE } keep_alive;
	
	int file_started;
	int file_finished;
	
	chunkqueue *write_queue;
	
	off_t bytes_written;          /* used by mod_accesslog */
	off_t bytes_written_last_sec; /* used by mod_status */
	off_t bytes_header;
	
	int http_status;
	
	sock_addr dst_addr;

	/* request */
	buffer *parse_request;
	unsigned int parsed_response; /* bitfield which contains the important header-fields of the parsed response header */
	
	request  request;
	request_uri uri;
	physical physical; 
	response response;
	
	size_t header_len;
	buffer *docroot_cache_key;
	buffer *docroot_cache_value;
	
	buffer *authed_user;
	
	
	/* response */
	int    got_response;
	
	int    in_joblist;
	
	connection_type mode;
	
	fcgi  fcgi;
	cgi   cgi;
	cache cache;
	
	file_cache_entry *fce;
	
#ifdef USE_OPENSSL
	SSL *ssl;
#endif
} connection;

typedef struct {
	connection **ptr;
	size_t size;
	size_t used;
} connections;


#ifdef HAVE_IPV6
typedef struct {
	int family;
	union {
		struct in6_addr ipv6;
		struct in_addr  ipv4;
	} addr;
	char b2[INET6_ADDRSTRLEN + 1];
	time_t ts;
} inet_ntop_cache_type;
#endif


typedef struct {
	buffer *uri;
	time_t mtime;
	int http_status;
} realpath_cache_type;

typedef struct {
	time_t  mtime;  /* the key */
	buffer *str;    /* a buffer for the string represenation */
} mtime_cache_type;

typedef struct {
	char *start;
	size_t len;
} dot;

typedef struct {
	dot **ptr;
	size_t used;
	size_t size;
} dot_stack;

typedef struct {
	void  *ptr;
	size_t used;
	size_t size;
} buffer_plugin;

typedef struct {
#ifdef HAVE_PCRE
	pcre *key;
#endif
	array *indexfiles;
	array *mimetypes;
	
	/* virtual-servers */
	buffer *document_root;
	
	unsigned short dump_unknown_headers;
	unsigned short dir_listing;
} specific_config;

typedef struct {
	specific_config **ptr;
	
	size_t used;
	size_t size;
} specific_config_array;

typedef struct {
	unsigned short port;
	buffer *bindhost;
	buffer *error_logfile;
	unsigned short dont_daemonize;
	unsigned short use_ipv6;
	buffer *changeroot;
	buffer *username;
	buffer *groupname;
	buffer *ssl_pemfile;
	
	buffer *errorfile_prefix;
	buffer *license;
	buffer *pid_file;
	
	buffer *event_handler;
	
	array *modules;
	
	unsigned short max_request_size;
} server_config;

typedef struct {
	sock_addr srv_addr;
	
	int fd;
	int log_error_fd;
	int log_using_syslog;
	fdevents *ev, *ev_ins;
	
	buffer_plugin plugins;
	void *plugin_slots;
	
	int con_opened;
	int con_read;
	int con_written;
	int con_closed;
	
	int is_ssl;
	
#ifdef USE_OPENSSL
	SSL_CTX *ssl_ctx;
#endif
	
	/* buffers */
	buffer *parse_full_path;
	buffer *response_header;
	buffer *error_log;
	buffer *response_range;
	buffer *tmp_buf;
	
	buffer *tmp_chunk_len;
	
	buffer *range_buf;
	
	/* caches */
#ifdef HAVE_IPV6
	inet_ntop_cache_type inet_ntop_cache[INET_NTOP_CACHE_MAX];
#endif
	mtime_cache_type mtime_cache[FILE_CACHE_MAX];
	
	dot_stack stack;
	
	/* Timestamps */
	time_t cur_ts;
	time_t last_generated_date_ts;
	time_t last_generated_debug_ts;
	
	buffer *ts_debug_str;
	buffer *ts_date_str;
	
	/* config-file */
	
	array *config;
	array *config_touched;
	
	array *config_context;
	
	specific_config_array *config_by_url;
	specific_config_array *config_by_host;
	specific_config_array *config_by_file;
	
	specific_config conf;
	server_config   srvconf;
	
	int config_deprecated;
	
	connections *conns;
	connections *joblist;
	
	file_cache  *file_cache;
	buffer      *file_cache_etag;
	
	fdevent_handler_t event_handler;
} server;


#endif
