#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#include <zlib.h>

#include "buffer.h"
#include "server.h"
#include "log.h"
#include "connections.h"
#include "fdevent.h"

#include "request.h"
#include "response.h"
#include "network.h"
#include "http_chunk.h"
#include "file_cache.h"

#include "plugin.h"

#ifdef USE_OPENSSL
# include <openssl/ssl.h> 
# include <openssl/err.h> 
#endif

#ifdef HAVE_SYS_FILIO_H
# include <sys/filio.h>
#endif

const char *connection_get_state(connection_state_t state) {
	switch (state) {
	case CONNECT: return "connect";
	case READ: return "read";
	case READ_POST: return "readpost";
	case WRITE: return "write";
	case CLOSE: return "close";
	case ERROR: return "error";
	case HANDLE_REQUEST: return "handle-req";
	case REQUEST_START: return "req-start";
	case REQUEST_END: return "req-end";
	case RESPONSE_START: return "resp-start";
	case RESPONSE_END: return "resp-end";
	default: return "(unknown)";	
	}
}

const char *connection_get_short_state(connection_state_t state) {
	switch (state) {
	case CONNECT: return ".";
	case READ: return "r";
	case READ_POST: return "R";
	case WRITE: return "W";
	case CLOSE: return "C";
	case ERROR: return "E";
	case HANDLE_REQUEST: return "h";
	case REQUEST_START: return "q";
	case REQUEST_END: return "Q";
	case RESPONSE_START: return "s";
	case RESPONSE_END: return "S";
	default: return "x";	
	}
}

static connection *connections_get_new_connection(server *srv) {
	connections *conns = srv->conns;
	size_t i;
	
	if (conns->size == 0) {
		conns->size = 128;
		conns->ptr = malloc(sizeof(*conns->ptr) * conns->size);
		for (i = 0; i < conns->size; i++) {
			conns->ptr[i] = connection_init(srv);
		}
	} else if (conns->size == conns->used) {
		conns->size += 128;
		conns->ptr = realloc(conns->ptr, sizeof(*conns->ptr) * conns->size);
		
		for (i = conns->used; i < conns->size; i++) {
			conns->ptr[i] = connection_init(srv);
		}
	}

	connection_reset(srv, conns->ptr[conns->used]);
#if 0	
	fprintf(stderr, "%s.%d: add: ", __FILE__, __LINE__);
	for (i = 0; i < conns->used + 1; i++) {
		fprintf(stderr, "%d ", conns->ptr[i]->fd);
	}
	fprintf(stderr, "\n");
#endif	
	
	conns->ptr[conns->used]->ndx = conns->used;
	return conns->ptr[conns->used++];
}

static int connection_del(server *srv, connection *con) {
	size_t i;
	connections *conns = srv->conns;
	connection *temp;
	
	if (con == NULL) return -1;
	
	if (-1 == con->ndx) return -1;
	
	i = con->ndx;
	
	/* not last element */
	
	if (i != conns->used - 1) {
		temp = conns->ptr[i];
		conns->ptr[i] = conns->ptr[conns->used - 1];
		conns->ptr[conns->used - 1] = temp;
		
		conns->ptr[i]->ndx = i;
		conns->ptr[conns->used - 1]->ndx = -1;
	}
	
	conns->used--;
	
	con->ndx = -1;
#if 0
	fprintf(stderr, "%s.%d: del: (%d)", __FILE__, __LINE__, conns->used);
	for (i = 0; i < conns->used; i++) {
		fprintf(stderr, "%d ", conns->ptr[i]->fd);
	}
	fprintf(stderr, "\n");
#endif	
	return 0;
}

int connection_close(server *srv, connection *con) {
#if 0		
	struct timeval tv;

	gettimeofday(&(tv), NULL);

	fprintf(stderr, "%d: %.5f ms\n", con->fd,
		((tv.tv_sec - con->start_tv.tv_sec) * 1000.0) + 
		((tv.tv_usec - con->start_tv.tv_usec) / 1000.0)
		);
#endif	
	
	con->fde_ndx = fdevent_event_del(srv->ev, con->fde_ndx, con->fd);
	fdevent_fdc_del(srv->ev->fdc, con->fd);
	
	if (close(con->fd)) {
		log_error_write(srv, __FILE__, __LINE__, "sdd",
				"close:", con->fd, strerror(errno));
	}
#if 0
	log_error_write(srv, __FILE__, __LINE__, "sd",
			"closed()", con->fd);
#endif
	
	
	connection_del(srv, con);
	connection_set_state(srv, con, CONNECT);
	
	return 0;
}

static int connection_handle_read(server *srv, connection *con) {
	int len;
	/* ah, here is something */
					
	con->is_readable = 0;
	/* for now, blocking-read */
	
	buffer_prepare_append(con->request.request, 4096);
	
	/* remove terminating \0 */
	if (con->request.request->used > 0) con->request.request->used--;
	
#ifdef USE_OPENSSL
	if (srv->is_ssl) {
		len = SSL_read(con->ssl, 
			       con->request.request->ptr + con->request.request->used, 
			       con->request.request->size - con->request.request->used - 2);
	} else {
		len = read(con->fd, 
			   con->request.request->ptr + con->request.request->used, 
			   con->request.request->size - con->request.request->used - 2);
	}
#else
	len = read(con->fd, 
		   con->request.request->ptr + con->request.request->used, 
		   con->request.request->size - con->request.request->used - 2);

#endif
	
	if (len < 0) {
#ifdef USE_OPENSSL
		if (srv->is_ssl) {
			int r;
			
			switch ((r = SSL_get_error(con->ssl, len))) {
			case SSL_ERROR_WANT_READ:
				return 0;
			default:
				log_error_write(srv, __FILE__, __LINE__, "sdds", "SSL:", 
						r, srv->is_ssl,
						ERR_error_string(ERR_get_error(), NULL));
				break;
			}
		} else {
			if (errno == EAGAIN) return 0;
		
			if (errno != ECONNRESET) {
				/* expected for keep-alive */
				log_error_write(srv, __FILE__, __LINE__, "ssd", "connection closed - read failed: ", strerror(errno), errno);
			}
		}
#else
		if (errno == EAGAIN) return 0;
		
		if (errno != ECONNRESET) {
			/* expected for keep-alive */
			log_error_write(srv, __FILE__, __LINE__, "ssd", "connection closed - read failed: ", strerror(errno), errno);
		}
#endif
		connection_set_state(srv, con, ERROR);
		return -1;
	} else if (len == 0) {
		/* the other end close the connection -> KEEP-ALIVE */
#if 0
		log_error_write(srv, __FILE__, __LINE__, "s",
				"connection closed: remote site closed unexpectedly");
#endif 
		connection_set_state(srv, con, ERROR);
		return -1;
	}
	
	/* check if the request is finished -> \r\n\r\n */
	con->request.request->used += len;
	
	/* terminate buffer */
	con->request.request->ptr[con->request.request->used] = '\0';
	con->request.request->used++;
	
#if 0
	log_error_write(srv, __FILE__, __LINE__, "bd", con->request.request, con->request.request->used);
#endif
	/* the request is finished */
	if (http_request_header_finished(srv, con)) {
		srv->con_read++;
		
		connection_set_state(srv, con, REQUEST_END);
	} else if (con->request.request->used > MAX_HTTP_REQUEST_HEADER) {
		/* not finished */
		
		/* the request header itself is larger then 
		 * our our max_request_size 
		 *   -> this is a DOS, kill the connection
		 */
		
		connection_set_state(srv, con, ERROR);
	}
	
	return 0;
}


static int connection_handle_read_post(server *srv, connection *con) {
	int len = 1;
	size_t to_read;
	/* ah, here is something */
	
	con->is_readable = 0;
	
	/* how much has to be read ? 
	 * con->request.content->ptr   ... our buffer for the whole request
	 * con->request.content->used  ... how much is used up to now
	 * con->request.content_length ... content 
	 * 
	 * con->request.content_length should be con->request.content->used
	 * 
	 */
	
	to_read = con->request.content_length - con->request.content->used;
	
#ifdef USE_OPENSSL
	if (srv->is_ssl) {
		len = SSL_read(con->ssl, 
			       con->request.content->ptr + con->request.content->used, 
			       to_read);
	} else {
		len = read(con->fd, 
			   con->request.content->ptr + con->request.content->used, 
			   to_read);
	}
#else
	len = read(con->fd, 
		   con->request.content->ptr + con->request.content->used, 
		   to_read);
	
#endif
	
	if (len == -1) {
#ifdef USE_OPENSSL
		if (srv->is_ssl) {
			int r;
			
			switch ((r = SSL_get_error(con->ssl, len))) {
			case SSL_ERROR_WANT_READ:
				return 0;
			default:
				log_error_write(srv, __FILE__, __LINE__, "sds", "SSL:", 
						r, 
						ERR_error_string(ERR_get_error(), NULL));
				break;
			}
		} else {
			log_error_write(srv, __FILE__, __LINE__, "ss", "connection closed: read failed: ", strerror(errno));
		}
#else
		log_error_write(srv, __FILE__, __LINE__, "ss", "connection closed: read failed: ", strerror(errno));
#endif
		connection_set_state(srv, con, ERROR);
		return -1;
			
	} else if (len == 0) {
		/* the other end close the connection */
		log_error_write(srv, __FILE__, __LINE__, "s",
				"connection closed: remote side closed unexpectedly");
		
		connection_set_state(srv, con, ERROR);
		return -1;
	}
	
	
	con->request.content->used += len;
	
	/* the request is finished */
	if (con->request.content->used >= con->request.content_length) {
		connection_set_state(srv, con, HANDLE_REQUEST);
	}
	
	return 0;
}

static int connection_handle_write_prepare(server *srv, connection *con) {
	struct stat st;
	int s_len;
	
	switch(con->mode) {
	case MODULE:
	case DIRECT:
		switch(con->http_status) {
		case 400: /* class: header + custom body */
		case 401:
		case 403:
		case 404:
		case 411:
		case 416:
		case 500:
		case 501:
		case 503:
		case 505: {
			con->file_finished = 1;
				
			/* rewrite the filename */
			
			buffer_copy_string(con->physical.path, srv->srvconf.errorfile_prefix->ptr);
			buffer_append_string(con->physical.path, get_http_status_body_name(con->http_status));
			
			if (-1 == (stat(con->physical.path->ptr, &st))) {
				buffer *b;
				
				buffer_reset(con->physical.path);
				
				b = chunkqueue_get_append_buffer(con->write_queue);
				
				/* build default error-page */
				buffer_copy_string(b, 
					      "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
					      "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
					      "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
					      "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
					      " <head>\n"
					      "  <title>");
				buffer_append_long(b, con->http_status);
				buffer_append_string(b, " - ");
				buffer_append_string(b, get_http_status_name(con->http_status));
				
				buffer_append_string(b,
					      "</title>\n"
					      " </head>\n"
					      " <body>\n"
					      "  <h1>");
				buffer_append_long(b, con->http_status);
				buffer_append_string(b, " - ");
				buffer_append_string(b, get_http_status_name(con->http_status));
				
				buffer_append_string(b,"</h1>\n" 
					      " </body>\n"
					      "</html>\n"
					      );
				
				response_header_insert(srv, con, "Content-Type", "text/html");
			} else {
				/* get content-type */
				size_t k;
				s_len = con->physical.path->used - 1;
				
				for (k = 0; k < srv->conf.mimetypes->used; k++) {
					data_string *ds = (data_string *)srv->conf.mimetypes->data[k];
					int ct_len = ds->key->used - 1;
					
					if (s_len < ct_len ||
					    ds->key->used == 0) continue;
					
					if (0 == strncmp(con->physical.path->ptr + s_len - ct_len, ds->key->ptr, ct_len)) {
						response_header_insert(srv, con, "Content-Type", ds->value->ptr);
						break;
					}
				}
				
				if (k == srv->conf.mimetypes->used) {
					/* the other end close the connection */
					log_error_write(srv, __FILE__, __LINE__, "s",
							"connection closed: no mimetypes found (please use the shipped config-file)");
					
					
					connection_set_state(srv, con, ERROR);
					return -1;
				}
			}
		}
			/* fall through */
			
		case 200: /* class: header + body */
			if (con->physical.path->used) {
				con->file_finished = 1;
				
				if (NULL == (con->fce = file_cache_get_entry(srv, con, con->physical.path, con->fce))) {
					log_error_write(srv, __FILE__, __LINE__, "sb",
							strerror(errno), con->physical.path);
					
					connection_set_state(srv, con, ERROR);
					
					return -1;
				}
				
				if (S_ISREG(con->fce->st.st_mode)) {
					if (con->request.http_method == HTTP_METHOD_GET) {
						http_chunk_append_file(srv, con, con->physical.path, 0, con->fce->st.st_size);
						con->response.content_length = http_chunkqueue_length(srv, con);
					} else if (con->request.http_method == HTTP_METHOD_HEAD) {
						con->response.content_length = con->fce->st.st_size;
					} else {
						/* should never happen */
						SEGFAULT();
					}
					
					http_response_write_header(srv, con, 
								   con->response.content_length, 
								   con->fce->st.st_mtime);
					
					
				} else {
					/* why the heck ? */
					
					log_error_write(srv, __FILE__, __LINE__, "s",
							"connection closed: no regular-file to send");
					
					connection_set_state(srv, con, ERROR);
				}
			} else {
				if (con->file_finished) {
					con->response.content_length = (ssize_t)http_chunkqueue_length(srv, con);
				}
				
				/* disable keep-alive if size-info for the body is missing */
				if ((con->parsed_response & HTTP_CONTENT_LENGTH) &&
				    ((con->response.transfer_encoding & HTTP_TRANSFER_ENCODING_CHUNKED) == 0)) {
					con->keep_alive = FALSE;
				}
				
				http_response_write_header(srv, con, 
							   con->response.content_length, 
							   0);
			}
			break;
			
		case 206: /* write_queue is already prepared */
			http_response_write_header(srv, con, 
						   con->response.content_length,
						   0);
			con->file_finished = 1;
			break;
			
		case 205: /* class: header only */
		case 301:
		case 304:
		default:
			http_response_write_header(srv, con, 0, 0);
			con->file_finished = 1;
			break;
		}
		
		break;
	case FASTCGI:
	case CGI:
		if (con->request.http_method == HTTP_METHOD_HEAD) {
			/* remove possible chunks */
			chunkqueue_reset(con->write_queue);
		}
		
		if (0 == (con->parsed_response & HTTP_CONNECTION)) {
			/* (f)cgi did'nt send Connection: header 
			 * 
			 * shall we ?
			 */
			if (((con->response.transfer_encoding & HTTP_TRANSFER_ENCODING_CHUNKED) == 0) &&
			    (con->parsed_response & HTTP_CONTENT_LENGTH) == 0) {
				/* without content_length, no keep-alive */
				
				con->keep_alive = FALSE;
			}
		} else {
			/* the fcgi sent a connection header, keep it */
			con->keep_alive = con->response.keep_alive;
		}
		
		con->response.content_length = (ssize_t)http_chunkqueue_length(srv, con);
		http_response_write_basic_header(srv, con);
			
		break;
	}
	
	return 0;
}

static int connection_handle_write(server *srv, connection *con) {
	switch(network_write_chunkqueue(srv, con, con->write_queue)) {
	case 0:
		if (con->file_finished) {
#if 0
			log_error_write(srv, __FILE__, __LINE__, "sd",
				"write done", con->fd);
#endif
			connection_set_state(srv, con, RESPONSE_END);
		}
		break;
	case -1:
		log_error_write(srv, __FILE__, __LINE__, "sd",
				"connection closed: write failed", con->fd);
		connection_set_state(srv, con, ERROR);
		break;
	case 1:
		con->is_writable = 0;
		
		/* not finished yet -> WRITE */
		break;
	}
	
	return 0;
}



connection *connection_init(server *srv) {
	connection *con;
	
	UNUSED(srv);

	con = calloc(1, sizeof(*con));
		
	con->fd = 0;
	con->ndx = -1;
	con->fde_ndx = -1;
	con->fcgi.fde_ndx = -1;
	con->bytes_written = 0;
	con->bytes_written_last_sec = 0;
	con->bytes_header = 0;
	

#define CLEAN(x) \
	con->x = buffer_init();
	
	CLEAN(request.uri);
	CLEAN(request.request_line);
	CLEAN(request.request);
	CLEAN(request.pathinfo);
	CLEAN(request.content);
	
	CLEAN(uri.scheme);
	CLEAN(uri.authority);
	CLEAN(uri.path);
	CLEAN(uri.path_raw);
	CLEAN(uri.query);
	
	CLEAN(physical.doc_root);
	CLEAN(physical.path);
	CLEAN(physical.etag);
	
	CLEAN(fcgi.response);
	CLEAN(fcgi.response_header);
	CLEAN(fcgi.write_buffer);
	CLEAN(cgi.response);
	CLEAN(cgi.response_header);
	
	CLEAN(parse_request);
	
	CLEAN(docroot_cache_key);
	CLEAN(docroot_cache_value);
	
	CLEAN(cache.trigger_handler);
	CLEAN(cache.basedir);
	CLEAN(cache.session_id);
	
	CLEAN(authed_user);
	
#undef CLEAN
	con->write_queue = chunkqueue_init();
	con->cache.trigger_if     = buffer_array_init();
	con->cache.output_include = buffer_array_init();
	con->cache.eval           = buffer_array_init();
	con->request.headers      = array_init();
	con->response.headers     = array_init();
	
	return con;
}

void connections_free(server *srv) {
	connections *conns = srv->conns;
	size_t i;	
	
	for (i = 0; i < conns->size; i++) {
		connection *con = conns->ptr[i];
		
		connection_reset(srv, con);
		
		chunkqueue_free(con->write_queue);
		buffer_array_free(con->cache.trigger_if);
		buffer_array_free(con->cache.output_include);
		buffer_array_free(con->cache.eval);
		array_free(con->request.headers);
		array_free(con->response.headers);

#define CLEAN(x) \
	buffer_free(con->x);
		
		CLEAN(request.uri);
		CLEAN(request.request_line);
		CLEAN(request.request);
		CLEAN(request.pathinfo);
		CLEAN(request.content);
		
		CLEAN(uri.scheme);
		CLEAN(uri.authority);
		CLEAN(uri.path);
		CLEAN(uri.path_raw);
		CLEAN(uri.query);
		
		CLEAN(physical.doc_root);
		CLEAN(physical.path);
		CLEAN(physical.etag);
		
		CLEAN(fcgi.response);
		CLEAN(fcgi.response_header);
		CLEAN(fcgi.write_buffer);
		
		CLEAN(cgi.response);
		CLEAN(cgi.response_header);
		
		CLEAN(parse_request);
		
		CLEAN(docroot_cache_key);
		CLEAN(docroot_cache_value);
		
		CLEAN(cache.trigger_handler);
		CLEAN(cache.basedir);
		CLEAN(cache.session_id);
		
		CLEAN(authed_user);
#undef CLEAN
		if (con->fcgi.rb) {
			if (con->fcgi.rb->ptr) free(con->fcgi.rb->ptr);
			free(con->fcgi.rb);
		}
		
		free(con);
	}
	
	free(conns->ptr);
}


int connection_reset(server *srv, connection *con) {
	con->is_readable = 1;
	con->is_writable = 1;
	con->http_status = 0;
	con->file_finished = 0;
	con->file_started = 0;
	con->got_response = 0;
	
	con->parsed_response = 0;
	
	con->bytes_written = 0;
	con->bytes_written_last_sec = 0;
	con->bytes_header = 0;
	
	con->request.http_method = HTTP_METHOD_UNSET;
	con->request.http_version = HTTP_VERSION_UNSET;
	
	con->request.http_if_modified_since = NULL;
	con->request.http_if_none_match = NULL;
	
	con->response.keep_alive = 0;
	con->response.content_length = -1;
	con->response.transfer_encoding = 0;
	
	con->cgi.pid = 0;
	con->fcgi.request_id = 0;
	con->fcgi.state = FCGI_STATE_INIT;
	con->fcgi.host = NULL;
	
	con->mode = DIRECT;
	
#define CLEAN(x) \
	if (con->x) buffer_reset(con->x);
	
	CLEAN(request.uri);
	CLEAN(request.request_line);
	CLEAN(request.pathinfo);
	CLEAN(request.content);
	
	CLEAN(uri.scheme);
	CLEAN(uri.authority);
	CLEAN(uri.path);
	CLEAN(uri.path_raw);
	CLEAN(uri.query);
	
	CLEAN(physical.doc_root);
	CLEAN(physical.path);
	CLEAN(physical.etag);
	
	CLEAN(fcgi.response);
	CLEAN(fcgi.response_header);
	CLEAN(fcgi.write_buffer);
	
	CLEAN(cgi.response);
	CLEAN(cgi.response_header);
	
	CLEAN(parse_request);
	
	CLEAN(cache.trigger_handler);
	CLEAN(cache.basedir);
	CLEAN(cache.session_id);
	
	CLEAN(authed_user);
#undef CLEAN	
	
#define CLEAN(x) \
	if (con->x) con->x->used = 0;	
	
	CLEAN(cache.trigger_if);
	CLEAN(cache.output_include);
	CLEAN(cache.eval);
#undef CLEAN
	
#define CLEAN(x) \
		con->request.x = NULL;
	
	CLEAN(http_host);
	CLEAN(http_range);
	CLEAN(http_content_type);
#undef CLEAN
	con->request.content_length = 0;
	
	array_reset(con->request.headers);
	array_reset(con->response.headers);
	
	chunkqueue_reset(con->write_queue);
	
	if (con->fce) {
		file_cache_entry_release(srv, con, con->fce);
		con->fce = NULL;
	}
	
	con->fcgi.response_len = 0;
	con->fcgi.response_type = 0;
	con->fcgi.response_padding = 0;
	con->fcgi.response_request_id = 0;
	con->fcgi.fd = -1;

#ifdef USE_OPENSSL
	if (srv->is_ssl) {
		if (con->ssl) SSL_free(con->ssl);
		if (NULL == (con->ssl = SSL_new(srv->ssl_ctx))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
					ERR_error_string(ERR_get_error(), NULL));
		} else {
			SSL_set_accept_state(con->ssl);
		}
	}
#endif
	
	if (con->response.keep_alive) {
		if (con->request.request->used) {
			size_t content_len = 0;
			if (HTTP_METHOD_POST == con->request.http_method) {
				content_len = con->request.content_length;
			}
			
			memmove(con->request.request->ptr, 
				con->request.request->ptr + con->header_len + content_len, 
				con->request.request->used - con->header_len - content_len);
			con->request.request->used -= con->header_len - content_len;
			con->request.request->ptr[con->request.request->used] = '\0';
#if 0
			if (con->request.request->used) {
				log_error_write(srv, __FILE__, __LINE__, "sd", "pipelined: ", con->request.request->used);
			}
#endif
		}
	} else {
		buffer_reset(con->request.request);
	}
	
	con->header_len = 0;
	
	return 0;
}

connection *connection_accept(server *srv) {
	int accepted_requests = 0;
/* accept everything */

	/* search an empty place */
	int cnt;
	sock_addr cnt_addr;
	socklen_t cnt_len;
	/* accept it and register the fd */
	
	cnt_len = sizeof(cnt_addr);
	
	if (-1 == (cnt = accept(srv->fd, (struct sockaddr *) &cnt_addr, &cnt_len))) {
		if (errno != EAGAIN) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "accept failed: ", strerror(errno));
		}
		return NULL;
	} else {
		connection *con;
		
		accepted_requests++;
		/* ok, we have the connection, register it */
#if 0
		log_error_write(srv, __FILE__, __LINE__, "sd",
				"appected()", cnt);
#endif
		srv->con_opened++;
		
		con = connections_get_new_connection(srv);
		
		con->fd = cnt;
		con->fde_ndx = -1;
#if 0		
		gettimeofday(&(con->start_tv), NULL);
#endif		
		fdevent_fdc_add(srv->ev->fdc, con->fd, con, FD_EVENT_TYPE_CONNECTION);
		
		connection_set_state(srv, con, REQUEST_START);
		
		con->connection_start = srv->cur_ts;
		con->dst_addr = cnt_addr;
		
		if (-1 == (fdevent_fcntl_set(srv->ev, con->fd))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
			return NULL;
		}
#ifdef USE_OPENSSL
		/* connect FD to SSL */
		if (srv->is_ssl) {
			if (1 != (SSL_set_fd(con->ssl, cnt))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
						ERR_error_string(ERR_get_error(), NULL));
				return NULL;
			}
		}
#endif
		return con;
	}
}

int connection_handle_fdevent(server *srv, connection *con, int revents) {
	if (revents & FDEVENT_IN) {
		con->is_readable = 1;
#if 0
		log_error_write(srv, __FILE__, __LINE__, "sd", "read-wait - done", con->fd);
#endif
	}
	if (revents & FDEVENT_OUT) {
		con->is_writable = 1;
		/* we don't need the event twice */
		
		
	}
	
	
	if (revents & ~(FDEVENT_IN | FDEVENT_OUT)) {
		/* looks like an error */
						
		/* FIXME: revents = 0x19 still means that we should read from the queue */
		if (revents & FDEVENT_HUP) {
			if (con->state == CLOSE) {
				con->close_timeout_ts = 0;
			} else {
				/* sigio reports the wrong event here
				 * 
				 * there was no HUP at all 
				 */
#ifdef USE_LINUX_SIGIO
				if (srv->ev->in_sigio == 1) {
					log_error_write(srv, __FILE__, __LINE__, "sd",
						"connection closed: poll() -> HUP", con->fd);
				} else {
					connection_set_state(srv, con, ERROR);
				}
#else
				connection_set_state(srv, con, ERROR);
#endif
				
			}
		} else if (revents & FDEVENT_ERR) {
#ifndef USE_LINUX_SIGIO
			log_error_write(srv, __FILE__, __LINE__, "sd",
					"connection closed: poll() -> ERR", con->fd);
#endif	
			connection_set_state(srv, con, ERROR);
		} else {
			log_error_write(srv, __FILE__, __LINE__, "sd",
					"connection closed: poll() -> ???", revents);
		} 
	}
	
	if ((con->state == READ ||
	     con->state == READ_POST) &&
	    con->is_readable) {
		con->read_idle_ts = srv->cur_ts;
		
		if (con->state == READ) {
			connection_handle_read(srv, con);
			
			if (con->state == READ) {
				con->is_readable = 0;
			}
			
		} else if (con->state == READ_POST) {
			connection_handle_read_post(srv, con);
			
			if (con->state == READ_POST) {
				con->is_readable = 0;
			}
		}
	}
	
	if (con->state == WRITE &&
	    !chunkqueue_is_empty(con->write_queue) &&
	    con->is_writable) {
		
		if (-1 == connection_handle_write(srv, con)) {
			connection_set_state(srv, con, ERROR);
			
			log_error_write(srv, __FILE__, __LINE__, "ds",
					con->fd,
					"handle write failed.");
		} else if (con->state == WRITE) {
			con->write_request_ts = srv->cur_ts;
		}
	}
	
	if (con->state == CLOSE) {
		/* flush the read buffers */
		int b;
		
		if (ioctl(con->fd, FIONREAD, &b)) {
			log_error_write(srv, __FILE__, __LINE__, "ss",
					"ioctl() failed", strerror(errno));
		}
		
		if (b > 0) {
			char buf[1024];
			log_error_write(srv, __FILE__, __LINE__, "sdd",
					"CLOSE-read()", con->fd, b);
			
			/* */
			read(con->fd, buf, sizeof(buf));
		} else {
			/* nothing to read */
			
			con->close_timeout_ts = 0;
		}
	}
	
	return 0;
	
}

int connection_set_state(server *srv, connection *con, connection_state_t state) {
	UNUSED(srv);

	con->state = state;
	
	return 0;
}


int connection_state_machine(server *srv, connection *con) {
	int done = 0, r;
#if 0
	fprintf(stderr, "3. s(%d): %s ",
		con->fd,
		connection_get_state(con->state));
#endif
	while (done == 0) {
		size_t ostate = con->state;
		int b;
		
		switch (con->state) {
		case REQUEST_START: /* transient */
			con->request_start = srv->cur_ts;
			con->read_idle_ts = srv->cur_ts;
			
			connection_set_state(srv, con, READ);
			
			/* pipelining */
			if (con->request.request->used && http_request_header_finished(srv, con)) {
				connection_set_state(srv, con, REQUEST_END);
			}
			
			break;
		case REQUEST_END: /* transient */
			if (http_request_parse(srv, con)) {
				/* we have to read some data from the POST request */
				connection_set_state(srv, con, READ_POST);
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sdd", "fd (read-post)", con->fd, con->state);
#endif
				break;
			}
			
			connection_set_state(srv, con, HANDLE_REQUEST);
			
			break;
		case HANDLE_REQUEST:
			/* 
			 * the request is parsed
			 * 
			 * decided what to do with the request
			 * - 
			 * 
			 * 
			 */
			
			switch (r = http_response_prepare(srv, con)) {
			case HANDLER_FINISHED:
				/* we have something to send, go on */
				connection_set_state(srv, con, RESPONSE_START);
				break;
			case HANDLER_COMEBACK:
				done = -1;
			case HANDLER_WAIT_FOR_EVENT:
				/* come back here */
				connection_set_state(srv, con, HANDLE_REQUEST);
				break;
			case HANDLER_ERROR:
				/* something went wrong */
				connection_set_state(srv, con, ERROR);
				break;
			default:
				log_error_write(srv, __FILE__, __LINE__, "sdd", "unknown ret-value: ", con->fd, r);
				break;
			}
			
			break;
		case RESPONSE_START:
			/* 
			 * the decision is done
			 * - create the HTTP-Response-Header
			 * 
			 */
			connection_handle_write_prepare(srv, con);
			
			connection_set_state(srv, con, WRITE);
			break;
		case RESPONSE_END: /* transient */
			/* log the request */
			
			plugins_call_handle_request_done(srv, con);
			
			srv->con_written++;
			
			connection_reset(srv, con);
			
			if (con->keep_alive) {
				connection_set_state(srv, con, REQUEST_START);
				
#if 0					
				con->request_start = srv->cur_ts;
				con->read_idle_ts = srv->cur_ts;
#endif
			} else {
				switch(r = plugins_call_handle_connection_close(srv, con)) {
				case HANDLER_GO_ON:
					break;
				default:
					log_error_write(srv, __FILE__, __LINE__, "");
					break;
				}
				
#ifdef USE_OPENSSL
				if (srv->is_ssl) {
					if (1 != (SSL_shutdown(con->ssl))) {
						log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:", 
								ERR_error_string(ERR_get_error(), NULL));
						return -1;
					}
				}
#endif
				

#if 0
				if (0 == shutdown(con->fd, SHUT_WR)) {
					con->close_timeout_ts = srv->cur_ts;
					connection_set_state(srv, con, CLOSE);
# if 1
					log_error_write(srv, __FILE__, __LINE__, "sd",
							"shutdown()", con->fd);
# endif
				} else {
					connection_close(srv, con);
					
# if 0
					log_error_write(srv, __FILE__, __LINE__, "sd",
							"closed()", con->fd);
# endif
				}
#else
				connection_close(srv, con);
#endif
				
				srv->con_closed++;
			}
			break;
		case CONNECT:
			buffer_reset(con->request.request);
			
			break;
		case CLOSE:
			if (con->keep_alive) {
				if (ioctl(con->fd, FIONREAD, &b)) {
					log_error_write(srv, __FILE__, __LINE__, "ss",
							"ioctl() failed", strerror(errno));
				}
				if (b > 0) {
					char buf[1024];
					log_error_write(srv, __FILE__, __LINE__, "sdd",
							"CLOSE-read()", con->fd, b);
					
					/* */
					read(con->fd, buf, sizeof(buf));
				} else {
					/* nothing to read */
					
					con->close_timeout_ts = 0;
				}
			} else {
				con->close_timeout_ts = 0;
			}
			
			if (srv->cur_ts - con->close_timeout_ts > 1) {
				connection_close(srv, con);
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sd",
						"closed()", con->fd);
#endif
			}
			
			break;
		case READ_POST:
		case READ:
			if (con->is_readable) {
				con->read_idle_ts = srv->cur_ts;
				
				if (con->state == READ) {
					connection_handle_read(srv, con);
					
					if (con->state == READ) {
						con->is_readable = 0;
					}
				} else if (con->state == READ_POST) {
					connection_handle_read_post(srv, con);
					
					if (con->state == READ_POST) {
						con->is_readable = 0;
					}
				}
			}
			break;
		case WRITE:
			/* only try to write if we have something in the queue */
			if (!chunkqueue_is_empty(con->write_queue)) {
#if 0
				log_error_write(srv, __FILE__, __LINE__, "dsd",
						con->fd,
						"packets to write:",
						con->write_queue->used);
#endif
			}
			if (!chunkqueue_is_empty(con->write_queue) && con->is_writable) {
				if (-1 == connection_handle_write(srv, con)) {
					log_error_write(srv, __FILE__, __LINE__, "ds",
							con->fd,
							"handle write failed.");
					connection_set_state(srv, con, ERROR);
				} else if (con->state == WRITE) {
					con->write_request_ts = srv->cur_ts;
				}
			}
			
			break;
		case ERROR: /* transient */
			
			/* even if the connection was drop we still have to write it to the access log */
			if (con->http_status) {
				plugins_call_handle_request_done(srv, con);
			}
#ifdef USE_OPENSSL
			if (srv->is_ssl) {
				int ret;
				switch ((ret = SSL_shutdown(con->ssl))) {
				case 1:
					/* ok */
					break;
				case 0:
					SSL_shutdown(con->ssl);
					break;
				default:
					log_error_write(srv, __FILE__, __LINE__, "sds", "SSL:", 
							SSL_get_error(con->ssl, ret), 
							ERR_error_string(ERR_get_error(), NULL));
					return -1;
				}
			}
#endif
			
			switch(con->mode) {
			case DIRECT:
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sd", 
						"emergency exit: direct", 
						con->fd);
#endif
				break;
			default:
				switch(r = plugins_call_handle_connection_close(srv, con)) {
				case HANDLER_GO_ON:
				case HANDLER_FINISHED:
					break;
				default:
					log_error_write(srv, __FILE__, __LINE__, "");
					break;
				}
				break;
			}
			
				
			connection_reset(srv, con);
			
			/* close the connection */
			if ((con->keep_alive == TRUE) &&
			    (0 == shutdown(con->fd, SHUT_WR))) {
				con->close_timeout_ts = srv->cur_ts;
				connection_set_state(srv, con, CLOSE);
#if 0
				log_error_write(srv, __FILE__, __LINE__, "sd",
						"shutdown()", con->fd);
#endif
			} else {
				connection_close(srv, con);
			}
			
			con->keep_alive = FALSE;
			
			srv->con_closed++;
			
			break;
		default:
			log_error_write(srv, __FILE__, __LINE__, "sdd", 
					"unknown state:", con->fd, con->state);
			
			break;
		}
		
		if (done == -1) {
			done = 0;
		} else if (ostate == con->state) {
			done = 1;
		}
	}
	
#if 0
	fprintf(stderr, "%s\n",
		get_state(con->state));
#endif

	return 0;
}
