#include <sys/types.h>

#include "settings.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#include "fdevent.h"

#include <fcntl.h>

fdevents *fdevent_init(size_t maxfds, fdevent_handler_t type) {
	fdevents *ev;
	size_t i;
	
	ev = calloc(1, sizeof(*ev));
	
	switch(type) {
	case FDEVENT_HANDLER_POLL:
		if (0 != fdevent_poll_init(ev, maxfds)) {
			fprintf(stderr, "%s.%d: event-handler poll failed\n", 
				__FILE__, __LINE__);
			
			return NULL;
		}
		break;
	case FDEVENT_HANDLER_SELECT:
		if (0 != fdevent_select_init(ev, maxfds)) {
			fprintf(stderr, "%s.%d: event-handler select failed\n", 
				__FILE__, __LINE__);
			return NULL;
		}
		break;
	case FDEVENT_HANDLER_LINUX_RTSIG:
		if (0 != fdevent_linux_rtsig_init(ev, maxfds)) {
			fprintf(stderr, "%s.%d: event-handler linux-rtsig failed, try to set server.event-handler = \"poll\" or \"select\"\n", 
				__FILE__, __LINE__);
			return NULL;
		}
		break;
	case FDEVENT_HANDLER_LINUX_SYSEPOLL:
		if (0 != fdevent_linux_sysepoll_init(ev, maxfds)) {
			fprintf(stderr, "%s.%d: event-handler linux-sysepoll failed, try to set server.event-handler = \"poll\" or \"select\"\n", 
				__FILE__, __LINE__);
			return NULL;
		}
		break;
	case FDEVENT_HANDLER_SOLARIS_DEVPOLL:
		if (0 != fdevent_solaris_devpoll_init(ev, maxfds)) {
			fprintf(stderr, "%s.%d: event-handler solaris-devpoll failed, try to set server.event-handler = \"poll\" or \"select\"\n", 
				__FILE__, __LINE__);
			return NULL;
		}
		break;
	case FDEVENT_HANDLER_FREEBSD_KQUEUE:
		if (0 != fdevent_freebsd_kqueue_init(ev, maxfds)) {
			fprintf(stderr, "%s.%d: event-handler freebsd-kqueue failed, try to set server.event-handler = \"poll\" or \"select\"\n", 
				__FILE__, __LINE__);
			return NULL;
		}
		break;
	default:
		fprintf(stderr, "%s.%d: event-handler is unknown, try to set server.event-handler = \"poll\" or \"select\"\n", 
			__FILE__, __LINE__);
		return NULL;
	}
	
	ev->fdc = calloc(1, sizeof(*ev->fdc));
	ev->fdc->size = maxfds * 2;
	ev->fdc->ptr = malloc(sizeof(*ev->fdc->ptr) * ev->fdc->size);
	for (i = 0; i < ev->fdc->size; i++) {
		ev->fdc->ptr[i].fd = -1;
	}
	
	return ev;
}

void fdevent_free(fdevents *ev) {
	if (!ev) return;
	
	if (ev->free) ev->free(ev);
	
	free(ev->fdc->ptr);
	free(ev->fdc);
	
	free(ev);
}

int fdevent_reset(fdevents *ev) {
	if (ev->reset) return ev->reset(ev);
	
	return 0;
}

int fdevent_event_del(fdevents *ev, int fde_ndx, int fd) {
	if (ev->event_del) return ev->event_del(ev, fde_ndx, fd);
	
	return -1;
}

int fdevent_event_add(fdevents *ev, int fde_ndx, int fd, int events) {
	if (ev->event_add) return ev->event_add(ev, fde_ndx, fd, events);
	return 0;
}

int fdevent_poll(fdevents *ev, int timeout_ms) {
	if (ev->poll) return ev->poll(ev, timeout_ms);
	return 0;
}

int fdevent_event_get_revent(fdevents *ev, size_t ndx) {
	if (ev->event_get_revent) return ev->event_get_revent(ev, ndx);
	return 0;
}

int fdevent_event_get_fd(fdevents *ev, size_t ndx) {
	if (ev->event_get_fd) return ev->event_get_fd(ev, ndx);
	return 0;
}

int fdevent_fcntl_set(fdevents *ev, int fd) {
	if (ev->fcntl_set) return ev->fcntl_set(ev, fd);
	
	return fcntl(fd, F_SETFL, O_NONBLOCK | O_RDWR);
}


int fdevent_event_next_fdndx(fdevents *ev, int ndx) {
	if (ev->event_next_fdndx) return ev->event_next_fdndx(ev, ndx);
	
	return -1;
}




/* FD -> Connection Mapping */

int fdevent_fdc_add(fd_conn_buffer *fdc, int fd, void *con, fd_event_t type) {
	if (fd < 0) return -1;
	
	if ((size_t)fd >= fdc->size) {
		return -1;
	}
	
	fdc->ptr[fd].fd = fd;
	fdc->ptr[fd].conn = con;
	fdc->ptr[fd].fd_type = type;
	
	fdc->used++;
	
	return 0;
}

int fdevent_fdc_get_index(fd_conn_buffer *fdc, int fd) {
	if (fdc->ptr[fd].fd == fd) return fd;
	
	return -1;
}

int fdevent_fdc_del(fd_conn_buffer *fdc, int fd) {
	if (fd < 0) return 0;
	
#if 0
	fprintf(stderr, "%s.%d: fdc-del: fd: %d\n", __FILE__, __LINE__, fd);
#endif	
#if 0
	if (fdc->ptr[fd].fd_type == FD_EVENT_TYPE_CGI_CONNECTION) {
		fprintf(stderr, "%s.%d: fdc-del: cgi-fd: %d\n", __FILE__, __LINE__, fd);
	}
#endif	
	if (fdc->ptr[fd].fd_type == FD_EVENT_TYPE_UNSET) {
		fprintf(stderr, "%s.%d: fd-type is not set for fd %d, dying here\n", __FILE__, __LINE__, fd);
		SEGFAULT();
	}
	
	fdc->ptr[fd].conn = NULL;
	fdc->ptr[fd].fd_type = FD_EVENT_TYPE_UNSET;
	fdc->ptr[fd].fd = -1;
	
	fdc->used--;
	
	return 0;
}
