#include <sys/types.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>

#include "fdevent.h"
#include "settings.h"

#ifdef USE_POLL
static void fdevent_poll_free(fdevents *ev) {
	free(ev->pollfds);
	free(ev->unused.ptr);
}

static int fdevent_poll_event_del(fdevents *ev, int fde_ndx, int fd) {
	if (fde_ndx < 0) return -1;
	
	if ((size_t)fde_ndx + 1 > ev->used) {
		fprintf(stderr, "%s.%d: del! out of range %d %u\n", __FILE__, __LINE__, fde_ndx, ev->used);
		SEGFAULT();
	}
	
	if (ev->pollfds[fde_ndx].fd == fd) {
		size_t k = fde_ndx;
		
		ev->pollfds[k].fd = -1;
		/* ev->pollfds[k].events = 0; */
		/* ev->pollfds[k].revents = 0; */
		
		if (ev->unused.size == 0) {
			ev->unused.size = 16;
			ev->unused.ptr = malloc(sizeof(*ev->unused.ptr) * ev->unused.size);
		} else if (ev->unused.size == ev->unused.used) {
			ev->unused.size += 16;
			ev->unused.ptr = realloc(ev->unused.ptr, sizeof(*ev->unused.ptr) * ev->unused.size);
		}
		
		ev->unused.ptr[ev->unused.used].ndx     = k;
		
		ev->unused.used++;
	} else {
		SEGFAULT();
	}
	
	return -1;
}

static int fdevent_poll_event_compress(fdevents *ev) {
	size_t j;
	int i;
	
	if (ev->unused.used == 0) return 0;
	
	for (j = 0; j < ev->unused.used; j++) {
		size_t k = ev->unused.ptr[j].ndx;
		size_t good_ndx;

		for (good_ndx = ev->used - 1; good_ndx + 1 > 0 && ev->pollfds[good_ndx].fd == -1; good_ndx--);
		
		if (good_ndx + 1 == 0) {
			SEGFAULT();
		}
		
		if (k < good_ndx) {
			connection       *con;
			file_cache_entry *fce;
			
			/* perhaps this section is a problem if the connection is gone while we have link to it */
			
			ev->pollfds[k].fd      = ev->pollfds[good_ndx].fd;
			ev->pollfds[k].events  = ev->pollfds[good_ndx].events;
			ev->pollfds[k].revents = ev->pollfds[good_ndx].revents;
			
			ev->pollfds[good_ndx].fd = -1;
#if 0
			fprintf(stderr, "(%d -> %d) ", good_ndx, k);
#endif			
			if (-1 == (i = fdevent_fdc_get_index(ev->fdc, ev->pollfds[k].fd))) {
				
				SEGFAULT();
			}
			
			con = ev->fdc->ptr[i].conn;
			fce = ev->fdc->ptr[i].conn;
				
			switch (ev->fdc->ptr[i].fd_type) {
			case FD_EVENT_TYPE_CONNECTION:
				if (con->fde_ndx != -1) con->fde_ndx = k;
				break;
			case FD_EVENT_TYPE_FCGI_CONNECTION:
				if (con->fcgi.fde_ndx != -1) con->fcgi.fde_ndx = k;
				break;
			case FD_EVENT_TYPE_CGI_CONNECTION:
				if (con->cgi.fde_ndx != -1) con->cgi.fde_ndx = k;
				break;
			case FD_EVENT_TYPE_DIRWATCH:
				if (fce->fde_ndx != -1) fce->fde_ndx = k;
				break;
			case -1:
				break;
			default:
				fprintf(stderr, "%s.%d: compress: (%d)\n", __FILE__, __LINE__, ev->pollfds[k].fd);
				SEGFAULT();
				break;
			}
		}
		
		ev->used--;
	}

	ev->unused.used = 0;
	
	return 0;
}

static int fdevent_poll_event_add(fdevents *ev, int fde_ndx, int fd, int events) {
	/* known index */
	if (fde_ndx != -1) {
		if (ev->pollfds[fde_ndx].fd == fd) {
			ev->pollfds[fde_ndx].events = events;
			
			return fde_ndx;
		}
		fprintf(stderr, "%s.%d: add: (%d, %d)\n", __FILE__, __LINE__, fde_ndx, ev->pollfds[fde_ndx].fd);
		SEGFAULT();
	}
	
	if (ev->size == 0) {
		ev->size = 16;
		ev->pollfds = malloc(sizeof(*ev->pollfds) * ev->size);
	} else if (ev->size == ev->used) {
		ev->size += 16;
		ev->pollfds = realloc(ev->pollfds, sizeof(*ev->pollfds) * ev->size);
	}
	
	ev->pollfds[ev->used].fd = fd;
	ev->pollfds[ev->used].events = events;
	
	return ev->used++;
}

static int fdevent_poll_poll(fdevents *ev, int timeout_ms) {
	fdevent_poll_event_compress(ev);
	
	return poll(ev->pollfds, ev->used, timeout_ms);
}

static int fdevent_poll_event_get_revent(fdevents *ev, size_t ndx) {
	if (ndx >= ev->used) {
		fprintf(stderr, "%s.%d: dying because: event: %u >= %u\n", __FILE__, __LINE__, ndx, ev->used);
		
		SEGFAULT();
		
		return 0;
	}
	return ev->pollfds[ndx].revents;
}

static int fdevent_poll_event_get_fd(fdevents *ev, size_t ndx) {
	return ev->pollfds[ndx].fd;
}

static int fdevent_poll_fcntl_set(fdevents *ev, int fd) {
	UNUSED(ev);

	return fcntl(fd, F_SETFL, O_NONBLOCK | O_RDWR);
}


static int fdevent_poll_event_next_fdndx(fdevents *ev, int ndx) {
	size_t i;
	
	i = (ndx < 0) ? 0 : ndx + 1;
	for (; i < ev->used; i++) {
		if (ev->pollfds[i].revents) break;
	}
	
	return i;
}

int fdevent_poll_init(fdevents *ev, size_t maxfds) {
	UNUSED(maxfds);

	ev->type = FDEVENT_HANDLER_POLL;
#define SET(x) \
	ev->x = fdevent_poll_##x;
	
	SET(free);
	SET(poll);
	
	SET(event_del);
	SET(event_add);
	
	SET(event_next_fdndx);
	SET(fcntl_set);
	SET(event_get_fd);
	SET(event_get_revent);
	
	return 0;
}

#else
int fdevent_poll_init(fdevents *ev, size_t maxfds) {
	fprintf(stderr, "%s.%d: poll is not supported, try to set server.event-handler = \"select\"\n",
			__FILE__, __LINE__);
	return -1;
}
#endif
